/*
 * This file is part of the Disco Deterministic Network Calculator v2.1.0 "Hydra".
 *
 * Copyright (C) 2005 - 2007 Frank A. Zdarsky
 * Copyright (C) 2011 - 2015 Steffen Bondorf
 *
 * disco | Distributed Computer Systems Lab
 * University of Kaiserslautern, Germany
 *
 * http://disco.cs.uni-kl.de
 *
 *
 * The Disco Deterministic Network Calculator (DiscoDNC) is free software;
 * you can redistribute it and/or modify it under the terms of the 
 * GNU Lesser General Public License as published by the Free Software Foundation; 
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

package unikl.disco.minplus;

import java.util.HashSet;
import java.util.Set;

import unikl.disco.curves.ArrivalCurve;
import unikl.disco.curves.Curve;
import unikl.disco.curves.ServiceCurve;
import unikl.disco.misc.Num;
import unikl.disco.nc.CalculatorConfig;

/**
 * 
 * @author Frank A. Zdarsky
 * @author Steffen Bondorf
 *
 */
public class Deconvolution {
	public static Set<ArrivalCurve> deconvolve( Set<ArrivalCurve> arrival_curves, Set<ServiceCurve> service_curves ) {
		Set<ArrivalCurve> results = new HashSet<ArrivalCurve>();
		
		for ( ServiceCurve beta : service_curves ) {
			for ( ArrivalCurve alpha : arrival_curves ) {
				results.add( deconvolve( alpha, beta ) );
			}
		}
		
		return results;
	}
	
	public static Set<ArrivalCurve> deconvolve_almostConcCs_SCs( Set<Curve> curves, Set<ServiceCurve> service_curves ) {
		Set<ArrivalCurve> results = new HashSet<ArrivalCurve>();
		
		for ( Curve sc : service_curves ) {
			for ( Curve c : curves ) {
				results.add( deconvolveGeneric( c, sc ) );
			}
		}
			
		return results;
	}

	public static Set<ArrivalCurve> deconvolve( Set<ArrivalCurve> arrival_curves, ServiceCurve service_curve ) {
		Set<ArrivalCurve> results = new HashSet<ArrivalCurve>();
		
		for ( ArrivalCurve alpha : arrival_curves ) {
			results.add( deconvolve( alpha, service_curve ) );
		}
		
		return results;
	}
	
	public static ArrivalCurve deconvolve( ArrivalCurve arrival_curve, ServiceCurve service_curve ) {
		return deconvolve( arrival_curve, service_curve, false );
	}
	
	public static ArrivalCurve deconvolve( ArrivalCurve arrival_curve, ServiceCurve service_curve, boolean tb_rl_optimized ) {
		if( service_curve.equals( ServiceCurve.createZeroDelayInfiniteBurst() ) ) {
			return arrival_curve;
		}
		if( service_curve.is_zero_delay_burst ) {
			return arrival_curve;
		}
		if( arrival_curve.equals( ArrivalCurve.createNullArrival() ) ) {
			return arrival_curve;
		}
		if( service_curve.equals( ServiceCurve.createNullService() )
				|| service_curve.getLatency().equals( Num.POSITIVE_INFINITY )
				|| ( service_curve.getSustainedRate().equals( Num.ZERO ) && service_curve.getSegment( service_curve.getSegmentCount() - 1 ).getY().equals( Num.ZERO ) ) ) {
			return ArrivalCurve.createNullArrival();
		}
		if ( tb_rl_optimized ) {
			return deconvolveTB_RL( arrival_curve, service_curve );
		} else {
			return deconvolveGeneric( arrival_curve, service_curve );
		}
	}
	
	public static ArrivalCurve deconvolveTB_RL( ArrivalCurve arrival_curve, ServiceCurve service_curve ) {
		if( service_curve.equals( ServiceCurve.createZeroDelayInfiniteBurst() ) ) {
			return arrival_curve;
		}
		if( service_curve.equals( ServiceCurve.createNullService() )
				|| service_curve.getLatency().equals( Num.POSITIVE_INFINITY )
				|| ( service_curve.getSustainedRate().equals( Num.ZERO ) && service_curve.getSegment( 1 ).getY().equals( Num.ZERO ) ) ) {
			return ArrivalCurve.createNullArrival();
		}
		
		return ArrivalCurve.createTokenBucket( arrival_curve.getSustainedRate().doubleValue(),
				arrival_curve.getBurst().doubleValue() + arrival_curve.getSustainedRate().doubleValue() * service_curve.getLatency().doubleValue() );
	}
	
	/**
	 * Returns the deconvolution of an (almost) concave arrival curve and
	 * a convex service curve.
	 * 
	 * @param curve_1 The (almost) concave arrival curve.
	 * @param curve_2 The convex service curve.
	 * @return The deconvolved curve, an arrival curve.
	 */
	private static ArrivalCurve deconvolveGeneric( Curve curve_1, Curve curve_2 ) {
		if( curve_2.getLatency().equals( Num.POSITIVE_INFINITY ) 
				|| ( curve_2.getSustainedRate().equals( Num.ZERO ) && curve_2.getSegment( curve_2.getSegmentCount() - 1 ).getY().equals( Num.ZERO ) ) ) {
			return ArrivalCurve.createNullArrival();
		}
		if ( CalculatorConfig.DECONVOLUTION_CHECKS ) {
			if( !curve_1.isAlmostConcave() ) {
				throw new IllegalArgumentException("Arrival curve of deconvolution must be almost concave.");
			}
			if ( !curve_2.isConvex() ) {
				throw new IllegalArgumentException("Service curve of deconvolution must be convex.");
			}
		}
		
		int i_this  = 0;
		int i_other = 0;

		// Advance this_iterator to concave part
		while(i_this < curve_1.getSegmentCount()) {
			if ( curve_1.getSegment( i_this ).getGrad().greater( Num.ZERO ) ) {
				break;
			}
			i_this++;
		}
		// Advance other_iterator to first IP after the start of this's concave part
		while(i_other < curve_2.getSegmentCount()) {
			if ( curve_2.getSegment( i_other ).getX().ge( curve_1.getSegment( i_this ).getX() ) ) {
				break;
			}
			i_other++;
		}

		// From here on, search both curve's IPs until grad(other)>=grad(this)
		Num kj = Num.POSITIVE_INFINITY;
		while(i_this < curve_1.getSegmentCount() || i_other < curve_2.getSegmentCount()) {
			Num x_this = ( i_this < curve_1.getSegmentCount() ) ?
					curve_1.getSegment( i_this ).getX() : Num.POSITIVE_INFINITY;
			Num x_other = (i_other < curve_2.getSegmentCount()) ?
					curve_2.getSegment(i_other).getX() : Num.POSITIVE_INFINITY;

			if ( x_other.le( x_this ) ) {
				if ( curve_2.getSegment(i_other).getGrad().ge
						( curve_1.getGradientLimitRight( curve_2.getSegment( i_other ).getX() ) ) ) {
					kj = curve_2.getSegment(i_other).getX();
					break;
				}
				i_other++;
			} else {
				if ( curve_2.getGradientLimitRight( curve_1.getSegment( i_this ).getX() ).ge
						( curve_1.getSegment( i_this ).getGrad() ) ) {
					kj = curve_1.getSegment(i_this).getX();
					break;
				}
				i_this++;
			}
		}

		if ( kj.less( Num.POSITIVE_INFINITY ) ) {
			ArrivalCurve result = new ArrivalCurve (Curve.add( Curve.shiftLeftClipping( curve_1.copy(), kj ), Num.negate( curve_2.f( kj ) ) ) );
			
			result.beautify();
			
			return result;
		} else {
			return ArrivalCurve.createZeroDelayInfiniteBurst();
		}
	}
}
